import pytest

from snakestream import Stream
from snakestream.collector import to_generator, to_list


@pytest.mark.asyncio
async def test_map(int_2_letter) -> None:
    # when
    it = Stream.of([1, 2, 3, 4]) \
        .map(lambda x: int_2_letter[x]) \
        .collect(to_generator)

    # then
    assert await it.__anext__() == 'a'
    assert await it.__anext__() == 'b'
    assert await it.__anext__() == 'c'
    assert await it.__anext__() == 'd'
    try:
        await it.__anext__()
    except StopAsyncIteration:
        pass
    else:
        assert False


@pytest.mark.asyncio
async def test_map_async_function(async_int_to_letter) -> None:
    async_int_to_letter = await async_int_to_letter

    # when
    it = Stream.of([1, 2, 3, 4]) \
        .map(async_int_to_letter) \
        .collect(to_generator)

    # then
    assert await it.__anext__() == 'a'
    assert await it.__anext__() == 'b'
    assert await it.__anext__() == 'c'
    assert await it.__anext__() == 'd'
    try:
        await it.__anext__()
    except StopAsyncIteration:
        pass
    else:
        assert False


@pytest.mark.asyncio
async def test_map_does_not_mutate_source(int_2_letter) -> None:
    source = [1, 2, 3, 4]

    # when
    it = await Stream.of(source) \
        .map(lambda x: int_2_letter[x]) \
        .collect(to_list)

    # then
    assert source != it
    assert len(source) == 4
    assert len(it) == 4




import pytest

from snakestream.stream import Stream


@pytest.mark.asyncio
async def test_is_parallel_parallel() -> None:
    # when
    res = Stream.of([1, 2, 3, 4, 1, 2, 3, 4]) \
        .parallel() \
        .distinct() \
        .is_parallel()
    # then
    assert res is True


@pytest.mark.asyncio
async def test_is_parallel_sequential() -> None:
    # when
    res = Stream.of([1, 2, 3, 4, 1, 2, 3, 4]) \
        .distinct() \
        .is_parallel()
    # then
    assert res is False




import pytest

from snakestream.collector import to_generator
from snakestream.stream import Stream


@pytest.mark.asyncio
async def test_iterate_simple() -> None:
    # when
    it = Stream.iterate(0, lambda n: n + 1) \
        .collect(to_generator)

    # then
    assert await it.__anext__() == 0
    assert await it.__anext__() == 1
    assert await it.__anext__() == 2
    assert await it.__anext__() == 3
    assert await it.__anext__() == 4
    assert await it.__anext__() == 5


@pytest.mark.asyncio
async def test_iterate_fib() -> None:
    # when
    it = Stream.iterate((0, 1), lambda n: (n[1], n[0] + n[1])) \
        .collect(to_generator)

    # then
    assert (await it.__anext__())[0] == 0
    assert (await it.__anext__())[0] == 1
    assert (await it.__anext__())[0] == 1
    assert (await it.__anext__())[0] == 2
    assert (await it.__anext__())[0] == 3
    assert (await it.__anext__())[0] == 5
    assert (await it.__anext__())[0] == 8
    assert (await it.__anext__())[0] == 13
    assert (await it.__anext__())[0] == 21
    assert (await it.__anext__())[0] == 34
    assert (await it.__anext__())[0] == 55



import pytest
import asyncio

from snakestream import Stream
from conftest import MyObject


@pytest.mark.asyncio
async def test_find_min_value_normal_input():
    input_list = [1, 2, 3, 4, 5]
    # when
    it = await Stream.of(input_list) \
        .min(lambda x, y: x > y)
    # then
    assert it == 1


@pytest.mark.asyncio
async def test_find_min_value_async_input():
    async def async_comparator(x: int, y: int) -> bool:
        await asyncio.sleep(0.01)
        return x > y

    input_list = [1, 2, 3, 4, 5]
    # when
    it = await Stream.of(input_list) \
        .min(async_comparator)
    # then
    assert it == 1


@pytest.mark.asyncio
async def test_find_min_value_empty_input():
    input_list = []
    # when
    it = await Stream.of(input_list) \
        .min(lambda x, y: x > y)
    # then
    assert it is None


@pytest.mark.asyncio
async def test_find_min_value_list_with_dupe_items():
    input_list = [1, 1, 2, 3, 4, 5]
    # when
    it = await Stream.of(input_list) \
        .min(lambda x, y: x > y)
    # then
    assert it == 1

    input_list = [1, 2, 3, 4, 5, 5]
    # when
    it = await Stream.of(input_list) \
        .min(lambda x, y: x > y)
    # then
    assert it == 1


@pytest.mark.asyncio
async def test_find_min_value_negative_values():
    input_list = [-1, -2, -3, -4, -5]
    # when
    it = await Stream.of(input_list) \
        .min(lambda x, y: x > y)
    # then
    assert it == -5


@pytest.mark.asyncio
async def test_find_min_value_custom_comparator():
    input_list = ['a', 'bb', 'ccc']
    # when
    it = await Stream.of(input_list) \
        .min(lambda x, y: len(x) > len(y))
    # then
    assert it == 'a'


@pytest.mark.asyncio
async def test_find_min_value_object_comparator() -> None:
    # when
    input_list = [MyObject(1, "object1"), MyObject(2, "object2"), MyObject(3, "object3"), MyObject(2, "object2"),
                  MyObject(3, "object3")]
    it = await Stream.of(input_list) \
        .min(lambda x, y: x.id > y.id)
    # then
    assert it == MyObject(1, "object1")



import pytest

from snakestream import Stream


@pytest.mark.asyncio
async def test_find_any() -> None:
    counter = 0

    def incr_counter(c):
        nonlocal counter
        counter += 1
        return c

    # when
    it = await Stream.of([1, 2, 3, 4, 5, 6, 1, 2, 3, 4, 5, 6]) \
        .map(incr_counter) \
        .filter(lambda x: x == 6) \
        .find_any()

    # then
    assert it == 6
    assert counter == 6


@pytest.mark.asyncio
async def test_find_any_found_none() -> None:
    counter = 0

    def incr_counter(c):
        nonlocal counter
        counter += 1
        return c

    # when
    it = await Stream.of([1, 2, 3, 4, 5, 6, 1, 2, 3, 4, 5, 6]) \
        .map(incr_counter) \
        .filter(lambda x: x == 100) \
        .find_any()

    # then
    assert it is None
    assert counter == 12


@pytest.mark.asyncio
async def test_find_any_empty_stream() -> None:
    counter = 0

    def incr_counter(c):
        nonlocal counter
        counter += 1
        return c

    # when
    it = await Stream.of([]) \
        .map(incr_counter) \
        .find_any()

    # then
    assert it is None
    assert counter == 0




import pytest
from snakestream.collector import to_list
from snakestream.stream import Stream


@pytest.mark.asyncio
async def test_limit_simple() -> None:
    # when
    lst = await Stream.iterate(0, lambda n: n + 1) \
        .limit(10) \
        .collect(to_list)

    # then
    assert len(lst) == 10


@pytest.mark.asyncio
async def test_limit_zero() -> None:
    # when
    lst = await Stream.iterate(0, lambda n: n + 1) \
        .limit(0) \
        .collect(to_list)

    # then
    assert len(lst) == 0


@pytest.mark.asyncio
async def test_limit_parallel() -> None:
    # when
    lst = await Stream.iterate(0, lambda n: n + 1) \
        .parallel() \
        .limit(10) \
        .collect(to_list)

    # then
    assert len(lst) == 10


@pytest.mark.asyncio
async def test_limit_multiple() -> None:
    # when
    lst = await Stream.of([[0, 1, 2], [3, 4], [5, 6, 7], [8, 9]]) \
        .limit(3) \
        .flat_map(lambda x: Stream.of(x)) \
        .limit(6) \
        .collect(to_list)

    # then
    assert lst == [0, 1, 2, 3, 4, 5]




import pytest

from snakestream.collector import to_generator
from snakestream.stream import Stream


@pytest.mark.asyncio
async def test_concat_simple() -> None:
    # when
    a = Stream.of([1, 2, 3, 4])
    b = Stream.of([5, 6, 7])

    generator = (await Stream.concat(a, b)) \
        .collect(to_generator)

    # then
    assert await generator.__anext__() == 1
    assert await generator.__anext__() == 2
    assert await generator.__anext__() == 3
    assert await generator.__anext__() == 4
    assert await generator.__anext__() == 5
    assert await generator.__anext__() == 6
    assert await generator.__anext__() == 7

    with pytest.raises(StopAsyncIteration):
        await generator.__anext__()


@pytest.mark.asyncio
async def test_concat_with_intermediaries() -> None:
    # when
    a = Stream.of([1, 2, 3, 4]) \
        .filter(lambda x: x < 3)
    b = Stream.of([5, 6, 7, 7]) \
        .distinct()

    generator = (await Stream.concat(a, b)) \
        .collect(to_generator)

    # then
    assert await generator.__anext__() == 1
    assert await generator.__anext__() == 2
    assert await generator.__anext__() == 5
    assert await generator.__anext__() == 6
    assert await generator.__anext__() == 7

    with pytest.raises(StopAsyncIteration):
        await generator.__anext__()




from contextlib import closing
import pytest

from snakestream.collector import to_list
from snakestream.stream import Stream


@pytest.mark.asyncio
async def test_close_simple(mocker, int_2_letter) -> None:
    mock_callback1 = mocker.Mock()
    mock_callback2 = mocker.Mock()

    stream = Stream.of([1, 2, 3, 4, 1, 2, 3, 4])

    it = await stream \
        .map(lambda x: int_2_letter[x]) \
        .distinct() \
        .on_close(mock_callback1) \
        .on_close(mock_callback2) \
        .collect(to_list)

    # when
    stream.close()

    # then
    mock_callback1.assert_called_once()
    mock_callback2.assert_called_once()

    assert 4 == len(it)
    assert 'a' in it
    assert 'b' in it
    assert 'c' in it
    assert 'd' in it


@pytest.mark.asyncio
async def test_close_after_stream_switch(mocker, int_2_letter) -> None:
    mock_callback1 = mocker.Mock()
    mock_callback2 = mocker.Mock()

    stream = Stream.of([1, 2, 3, 4, 1, 2, 3, 4])

    await stream \
        .map(lambda x: int_2_letter[x]) \
        .on_close(mock_callback1) \
        .distinct() \
        .parallel() \
        .on_close(mock_callback2) \
        .collect(to_list)

    # when
    stream.close()

    # then
    mock_callback1.assert_called_once()
    mock_callback2.assert_called_once()


@pytest.mark.asyncio
async def test_autoclose_simple(mocker, monkeypatch, int_2_letter):
    # given
    stream = Stream.of([1, 2, 3, 4, 1, 2, 3, 4])
    close_mock = mocker.Mock()
    monkeypatch.setattr(stream, 'close', close_mock)

    # when
    with closing(stream) as stream:
        it = await stream \
            .map(lambda x: int_2_letter[x]) \
            .distinct() \
            .collect(to_list)

    # then
    close_mock.assert_called_once()
    assert 4 == len(it)
